@@ -49,6 +49,7 @@
 void incrementDocRef(xmlDocPtr doc);
 int getNodeCount(xmlNodePtr node);
 void incrementDocRefBy(xmlDocPtr doc, int num);
+SEXP R_createXMLNodeRefDirect(xmlNodePtr node, int addFinalizer);
 
 /**
  Create a libxml comment node and return it as an S object
@@ -121,7 +122,7 @@ R_newXMLPINode(USER_OBJECT_ sdoc, USER_OBJECT_ name, USER_OBJECT_ content)
 
 USER_OBJECT_
 R_newXMLNode(USER_OBJECT_ name, USER_OBJECT_ attrs, USER_OBJECT_ nameSpace, USER_OBJECT_ sdoc,
-              USER_OBJECT_ nameSpaceDefinitions)
+              USER_OBJECT_ nameSpaceDefinitions, USER_OBJECT_ manageMemory)
 {
    xmlDocPtr doc = NULL;
    xmlNsPtr ns = NULL;
@@ -172,7 +173,7 @@ R_newXMLNode(USER_OBJECT_ name, USER_OBJECT_ attrs, USER_OBJECT_ nameSpace, USER
        XML_ROOT(doc) = node;
    }
 
-   return(R_createXMLNodeRef(node));
+   return(INTEGER(manageMemory)[0] > 0 ? R_createXMLNodeRef(node) : R_createXMLNodeRefDirect(node, 0));
 }
 
 USER_OBJECT_
@@ -551,9 +552,13 @@ R_insertXMLNode(USER_OBJECT_ node, USER_OBJECT_ parent, USER_OBJECT_ at, USER_OB
 	    /* tmp = xmlCopyNode(n, 1); */
 	} else {
 	    tmp = n;
-#ifdef R_XML_DEBUG
+
             if(n->_private)
+#ifdef R_XML_DEBUG
   	       fprintf(stderr, "insertXMLNode: %p incrementing document (%p)  %d\n", n, p->doc, *(int *) n->_private);
+	    if(strcmp(n->name, "table") == 0) {
+		fprintf(stderr, "table\n");
+	    }
 #endif
 	    incrementDocRefBy(p->doc, getNodeCount(n));
 	}
@@ -623,6 +628,12 @@ RS_XML_xmlAddSiblingAt(USER_OBJECT_ r_to, USER_OBJECT_ r_node, USER_OBJECT_ r_af
 	
     f = LOGICAL(r_after)[0] ?  xmlAddNextSibling : xmlAddPrevSibling ;
     ans = f(p, n);
+
+    /* If adding to the root node and inserting a node before the
+     * current first child, update the document.*/
+    if(p->doc && p->doc->children == p && n->next == p)
+	p->doc->children = n;
+
     incrementDocRefBy(p->doc, getNodeCount(n));
     return(R_createXMLNodeRef(ans));
 }
@@ -705,7 +716,7 @@ R_xmlRootNode(USER_OBJECT_ sdoc, USER_OBJECT_ skipDtd)
   }
 
   if(LOGICAL(skipDtd)[0]) {
-      while(node && (node->type == XML_DTD_NODE || node->type == XML_COMMENT_NODE)) {
+      while(node && node->type != XML_ELEMENT_NODE /* (node->type == XML_DTD_NODE || node->type == XML_COMMENT_NODE) */) {
 	  node = node->next;
       }
   }
@@ -857,14 +868,30 @@ RS_XML_clone(USER_OBJECT_ obj, USER_OBJECT_ recursive, USER_OBJECT_ addFinalizer
 xmlDocPtr currentDoc;
 #endif
 
-void incrementDocRefBy(xmlDocPtr doc, int num)
+int R_XML_MemoryMgrMarker = 1010101011;
+
+void
+initDocRefCounter(xmlDocPtr doc)
 {
     int *val;
-    if(!doc)
+    if(doc->_private)
 	return;
 
-    if(!doc->_private)
-	doc->_private = calloc(1, sizeof(int));
+    doc->_private = calloc(2, sizeof(int));
+    val = (int *) doc->_private;
+    val[1] = R_MEMORY_MANAGER_MARKER;
+}
+
+void 
+incrementDocRefBy(xmlDocPtr doc, int num)
+{
+    int *val;
+    if(!doc || IS_NOT_OUR_DOC_TO_TOUCH(doc))
+	return;
+
+    if(!doc->_private) {
+	initDocRefCounter(doc);
+    }
 
     val = (int *) doc->_private;
 
@@ -886,7 +913,8 @@ int getNodeCount(xmlNodePtr node)
 {
   int val = 0;
   xmlNodePtr p = node->children;
-  if(!node)
+
+  if(!node || IS_NOT_OUR_NODE_TO_TOUCH(node))
     return(0);
 
   val = GET_NODE_COUNT(node);
@@ -907,6 +935,10 @@ R_createXMLDocRef(xmlDocPtr doc)
   currentDoc = doc;
 #endif
 
+  if(!doc)
+     return(R_NilValue);
+
+  initDocRefCounter(doc);
   incrementDocRef(doc);
 
 #ifdef R_XML_DEBUG
@@ -966,6 +998,16 @@ R_convertXMLNsRef(SEXP r_ns)
   return(ans);
 }
 
+USER_OBJECT_
+R_getXMLNsRef(USER_OBJECT_ r_node)
+{
+    xmlNodePtr node = (xmlNodePtr) R_ExternalPtrAddr(r_node);
+
+    if(!node) 
+	return(R_NilValue);
+
+    return(node->ns ? R_createXMLNsRef(node->ns) : R_NilValue);
+}
 
 
 
@@ -1025,7 +1067,9 @@ R_getInternalNodeClass(xmlElementType type)
         case XML_DOCUMENT_FRAG_NODE:
               p = "XMLDocumentFragNode";
               break;
-
+        case XML_ATTRIBUTE_NODE:
+              p = "XMLAttributeNode";
+              break;
         default:
               p = "XMLUnknownInternalNode";
     }
@@ -1037,7 +1081,7 @@ void
 internal_incrementNodeRefCount(xmlNodePtr node)
 {
     int *val;
-    if(!node || !node->_private)
+    if(!node || IS_NOT_OUR_NODE_TO_TOUCH(node) || !node->_private)
 	return;
     val = (int *) node->_private;
     (*val)++;
@@ -1047,16 +1091,16 @@ SEXP
 R_getXMLRefCount(SEXP rnode)
 {
     xmlNodePtr node = (xmlNodePtr) R_ExternalPtrAddr(rnode);
-    if(!node || !node->_private)
+    if(!node || IS_NOT_OUR_NODE_TO_TOUCH(node) || !node->_private)
 	return(R_NilValue);
     return(ScalarInteger(*((int *) node->_private)));
 }
 
 int
-checkDescendantsInR(xmlNodePtr node)
+checkDescendantsInR(xmlNodePtr node, int process)
 {
     xmlNodePtr p;
-    if(!node)
+    if(!node && (process || IS_NOT_OUR_NODE_TO_TOUCH(node)))
 	return(0);
 
     if(node->_private)
@@ -1064,7 +1108,7 @@ checkDescendantsInR(xmlNodePtr node)
 
     p = node->children;
     while(p) {
-	if(checkDescendantsInR(p))
+	if(checkDescendantsInR(p, 0))
 	    return(1);
 	p = p->next;
     }
@@ -1076,7 +1120,8 @@ internal_decrementNodeRefCount(xmlNodePtr node)
 {
     int *val;
        /* */    
-    if(!node || !node->_private) /* if node->_private == NULL, should
+    if(!node || IS_NOT_OUR_NODE_TO_TOUCH(node))
+                                 /* if node->_private == NULL, should
 				  * we free this node?, i.e. if it is
 				  * not in a parent or a document.
                                   No! Basically we shouldn't get here
@@ -1084,6 +1129,9 @@ internal_decrementNodeRefCount(xmlNodePtr node)
                                   set the finalizer having set the _private */
 	return;
 
+    if(!node->_private)
+	return;
+
 
     /*  Get the value of the reference count and decrement it by 1.
         If we are now at 0, then we can potentially free this node.
@@ -1107,9 +1155,10 @@ internal_decrementNodeRefCount(xmlNodePtr node)
     fprintf(stderr, "decremented node (%s, %d) to %d  (%p)\n", node->name, node->type, *val, node);fflush(stderr);
 #endif
     if(*val == 0) {
+
 	free(node->_private);
         node->_private = NULL;
-	if(node->doc) {
+	if(node->doc && !IS_NOT_OUR_DOC_TO_TOUCH(node->doc)) {
 	    val = (int *) node->doc->_private;
 	    if(val) (*val)--;
 	    if(!val || *val == 0) {
@@ -1119,12 +1168,13 @@ internal_decrementNodeRefCount(xmlNodePtr node)
 #endif
 		if(val) free(node->doc->_private);
 		node->doc->_private = NULL;
+/*		fprintf(stderr, "<r:freeDoc addr='%p'/>\n", node->doc); */
 		xmlFreeDoc(node->doc);
 		R_numXMLDocsFreed++;
 	    }
 	} else if(!node->parent) {
             int hold;
-	    hold = checkDescendantsInR(node);
+	    hold = checkDescendantsInR(node, 1);
             if(!hold)
    	       xmlFreeNode(node);
 	} else {
@@ -1141,7 +1191,7 @@ internal_decrementNodeRefCount(xmlNodePtr node)
             xmlNodePtr p = node->parent;
 	    while(p->parent)  p = p->parent;
 
-	    hold = checkDescendantsInR(p);
+	    hold = checkDescendantsInR(p, 0);
 	    if(!hold)
 		xmlFree(p);
         
@@ -1159,34 +1209,19 @@ decrementNodeRefCount(SEXP rnode)
 }
 
 
-
-
-/**
-
- */
-USER_OBJECT_
-R_createXMLNodeRef(xmlNodePtr node)
+SEXP
+R_createXMLNodeRefDirect(xmlNodePtr node, int addFinalizer)
 {
   SEXP ref, tmp;
-  int *val;
-
-  if(!node)
-      return(NULL_USER_OBJECT);
-
-
-  if(node->_private == NULL) 
-      node->_private = calloc(1, sizeof(int));
-  val = (int *) node->_private;
-  (*val)++;
-  if(*val == 1)
-     incrementDocRef(node->doc);
-#ifdef R_XML_DEBUG
-  fprintf(stderr, "creating reference to node (%s, %d) count = %d (%p) (doc = %p)\n", node->name, node->type, (int) *val, node, node->doc);
-#endif
 
   PROTECT(ref = R_MakeExternalPtr(node, Rf_install("XMLInternalNode"), R_NilValue));
 #ifdef XML_REF_COUNT_NODES
-  R_RegisterCFinalizer(ref, decrementNodeRefCount);
+#ifdef XML_WITH_DEBUG
+fprintf(stderr, "Creating reference with finalizer for %s (%p) '%s'\n", 
+            node->name, node, node->type == XML_TEXT_NODE ? node->content : "");fflush(stderr);
+#endif
+  if(addFinalizer > 0 || (addFinalizer < 0 && !IS_NOT_OUR_NODE_TO_TOUCH(node)))
+     R_RegisterCFinalizer(ref, decrementNodeRefCount);
 #endif
   PROTECT(tmp = NEW_CHARACTER(3));
   SET_STRING_ELT(tmp, 0, mkChar(R_getInternalNodeClass(node->type)));
@@ -1197,6 +1232,74 @@ R_createXMLNodeRef(xmlNodePtr node)
   return(ref);
 }
 
+int
+clearNodeMemoryManagement(xmlNodePtr node)
+{
+    xmlNodePtr tmp;
+    int ctr = 0;
+    if(node->_private) {
+	free(node->_private);
+	node->_private = NULL;
+	ctr++;
+    }
+
+    tmp = node->children;
+    while(tmp) {
+	tmp = tmp->next;
+        if(tmp)
+   	   ctr += clearNodeMemoryManagement(tmp);
+    }
+    return(ctr);
+}
+
+SEXP
+R_clearNodeMemoryManagement(SEXP r_node)
+{
+   xmlNodePtr node = (xmlNodePtr) R_ExternalPtrAddr(r_node);
+   int val;
+
+   if(!node) 
+       return(ScalarInteger(-1));
+
+   val = clearNodeMemoryManagement(node);
+
+   return(ScalarInteger(val));
+}
+
+
+
+/**
+
+ */
+USER_OBJECT_
+R_createXMLNodeRef(xmlNodePtr node)
+{
+  int *val;
+
+  if(!node)
+      return(NULL_USER_OBJECT);
+
+/*  !IS_NOT_OUR_NODE_TO_TOUCH(node) */
+  if((node->_private && ((int*)node->_private)[1] == (int) R_MEMORY_MANAGER_MARKER)
+     || !node->doc || (!(IS_NOT_OUR_DOC_TO_TOUCH(node->doc)))) {
+      if(node->_private == NULL) {
+        node->_private = calloc(2, sizeof(int));
+	val = (int *) node->_private;
+	val[1] = R_MEMORY_MANAGER_MARKER;
+      }
+
+      val = (int *) node->_private;
+      (*val)++;
+      if(*val == 1)
+	  incrementDocRef(node->doc);
+#ifdef R_XML_DEBUG
+  fprintf(stderr, "creating reference to node (%s, %d) count = %d (%p) (doc = %p count = %d)\n", node->name, node->type, (int) *val, node, node->doc,  (node->doc && node->doc->_private) ? ((int *)node->doc->_private)[0] : -1);
+#endif
+  }
+
+  return(R_createXMLNodeRefDirect(node, !IS_NOT_OUR_NODE_TO_TOUCH(node)));
+}
+
 
 /*
  May not be used. Not yet.
@@ -1207,7 +1310,7 @@ SEXP
 R_addXMLNodeFinalizer(SEXP r_node)
 {
    xmlNodePtr node = (xmlNodePtr) R_ExternalPtrAddr(r_node);
-#ifdef XML_REF_COUNT_NODES
+#ifdef XML_REF_COUNT_NODES /* ??? should this be ifndef or ifdef.??  */
   R_RegisterCFinalizer(r_node, decrementNodeRefCount);
 #endif
   return(r_node);
@@ -1253,6 +1356,9 @@ R_saveXMLDOM(USER_OBJECT_ sdoc, USER_OBJECT_ sfileName, USER_OBJECT_ compression
 
     doc = (xmlDocPtr) R_ExternalPtrAddr(sdoc);
 
+    if(doc == NULL)
+	return(NEW_CHARACTER(0));
+
     xmlIndentTreeOutput = LOGICAL_DATA(sindent)[0];
 
     if(GET_LENGTH(prefix) == 3) {
@@ -1490,6 +1596,19 @@ RS_XML_printXMLNode(USER_OBJECT_ r_node, USER_OBJECT_ level, USER_OBJECT_ format
 }
 
 SEXP
+R_setXMLInternalTextNode_noenc(SEXP node)
+{
+     xmlNodePtr n = (xmlNodePtr) R_ExternalPtrAddr(node);
+     if(!n) {
+	 PROBLEM "null value passed for XMLInternalTextNode"
+	     ERROR;
+     }
+     n->name = (const xmlChar *) (&xmlStringTextNoenc);
+     return(ScalarLogical(TRUE));
+}
+
+SEXP
+/*R_setXMLInternalTextNode_value(SEXP node, SEXP value, SEXP r_encoding)*/
 R_setXMLInternalTextNode_value(SEXP node, SEXP value)
 {
    xmlNodePtr n = (xmlNodePtr) R_ExternalPtrAddr(node);
@@ -1514,7 +1633,7 @@ R_setXMLInternalTextNode_value(SEXP node, SEXP value)
 }
 
 SEXP
-R_xmlNodeValue(SEXP node, SEXP raw)
+R_xmlNodeValue(SEXP node, SEXP raw, SEXP r_encoding)
 {
    xmlNodePtr n = (xmlNodePtr) R_ExternalPtrAddr(node);
    xmlChar *tmp;
@@ -1538,7 +1657,13 @@ R_xmlNodeValue(SEXP node, SEXP raw)
    }
 */
    if(tmp) {
+#if 0
      ans = ScalarString(CreateCharSexpWithEncoding(encoding, tmp));
+#else
+     ans = ScalarString(mkCharCE(tmp, INTEGER(r_encoding)[0]));
+#endif
+
+     free(tmp);
 //     ans = mkString(XMLCHAR_TO_CHAR(tmp));
      // Just playing:  ans = ScalarString(mkCharCE(tmp, CE_UTF8));
    } else 
